--	SMH - Simply Magical Healer
--	Backend
--	Written by scott@mithrandir.com
--	Official website:  http://www.mithrandir.com/WorldOfWarcraft/WoW.html
--
--	This library keeps track of spells, talents and equipment so it
--	can provide information to determine the most appropriate healing
--	spell to use.  It uses a highly accurate +healing computation when
--	determining spell selection.
--
--	TODO:
--		* Casting times when initially read do not seem to have the modification
--		  from the talents.  Therefore, the cached data will not have the proper
--		  times if there are talents that change them.  However, when going to
--		  another instance will cause the spell book data to be reread and the
--		  talent-modified times will be gotten.  Note that when one manually
--		  examines the spell book, the talent casting times are present.  Also
--		  note that the mana cost reduction and the healing percentage increase
--		  seem to be present already in the cache.  It seems only the casting
--		  times are the problem.
--		* Need to correct the way the Ascendance buff is handled because a second
--		  one seems to contain the actual +healing value used.
--		* Determine if +350 is the proper value for the Dreamwalker benefit.
--		* Determine if Empowered Rejuvenation affects Lifebloom
--		* TBC: Deal with new talents that affect healing.
--			Druid:  Tree of Life -- assume a buff that indicates the amount of +healing ... party members get this
--				Tree of Life is a pain because it does not indicate +healing value, but 25% of associated Tree of Life's spirit.  What happens
--				when more than one is active?  Is there any way to record who the Tree of Life is and therefore know what the Spirit amount is?
--				This would require some interesting messaging to occur.  Especially when the Tree of Life's spirit changes for whatever reason.
--				When more than one Tree of Life is in a party each recipient has a buff from each Tree of Life.  The wording on the buffs is the
--				same, and I do not think there is a way to know which Tree is giving the buff.  Therefore, we need to look at each Tree in the
--				party, which is not that bad.
--			Paladin: Light's Grace (3): cast time of Holy Light reduced by 0.5 seconds  *** the basic code is in place, but we need to know if there are 3 ranks for 100% ***
--			Shaman: Healing Way (3):  Who on earth knows what this is doing?
--		* TBC: Deal with new items that affect healing.
--			Well Fed:  Increase Healing done by 44 ...    It seems one of the Well Fed gives you better healing.  Will need to parse it to find out.
--		* TBC: Deal with sockets, etc. and how they change +healing.
--		* Think about having the HOT amounts separated out from the base amounts
--		  so they can be examined/reported separately.
--		* Fecundity on the recipient is +30 healing, but it can stack up to 5 times.  How is that implemented?
--		* Look at Avatar Raiment Set for the new working for the extension of Renew (which one assumes changes heal amount)

SMH_Backend_File_Version = 09980

if not SMH or SMH.addonVersionNumber < SMH_Backend_File_Version then

	SMH_E = {			-- these are English language strings that are not to be localized
		DRUID   = 'DRUID',
		PALADIN = 'PALADIN',
		PRIEST  = 'PRIEST',
		SHAMAN  = 'SHAMAN',

		NOTIFICATION_BUFF_MODIFICATION_CHANGED               = 'SMHNOT_CHANGE_BUFF',
		NOTIFICATION_EQUIPMENT_MODIFICATION_CHANGED          = 'SMHNOT_CHANGE_EQUIP',
		NOTIFICATION_SPIRITUAL_GUIDANCE_MODIFICATION_CHANGED = 'SMHNOT_CHANGE_SG',
		NOTIFICATION_NATURES_BLESSING_MODIFICATION_CHANGED   = 'SMHNOT_CHANGE_NB',
		NOTIFICATION_HOLY_GUIDANCE_MODIFICATION_CHANGED      = 'SMHNOT_CHANGE_HG',
		NOTIFICATION_LUNAR_GUIDANCE_MODIFICATION_CHANGED     = 'SMHNOT_CHANGE_LG',
		NOTIFICATION_NURTURING_INSTINCT_MODIFICATION_CHANGED = 'SMHNOT_CHANGE_NI',
		NOTIFICATION_CACHED_SPELLBOOK                        = 'SMHNOT_CACHED_SPELLBOOK',
		}

	SMH = {

		addonVersionNumber = SMH_Backend_File_Version,

		buffsCancellingCasting = { [SMH_S.BAT_COSTUME] = true, [SMH_S.SNOWMAN] = true, [SMH_S.WISP_COSTUME] = true, },
		buffCasterHealingEffects = {
			[SMH_S.POWER_INFUSION] = 20,
			[SMH_S.CARRION_SWARM] = -75,
			[SMH_S.CHOSEN_ONE] = 100,
			[SMH_S.VILE_SLIME] = -50,
			[SMH_S.VILE_SLUDGE] = -50,
			},
		buffCasterPlusHealingEffects = {
			[SMH_S.BAND_OF_THE_ETERNAL_RESTORER] = 175,
			[SMH_S.BAND_OF_THE_ETERNAL_SAGE] = 95,
			[SMH_S.BLESSING_OF_ELUNE] = 210,
			[SMH_S.BLESSING_OF_RIGHTEOUSNESS] = 290,
			[SMH_S.BLESSING_OF_SILVER_CRESCENT] = 155,
			[SMH_S.BLOODTHISTLE] = 10,
			[SMH_S.COMBAT_GALLANTRY] = 80,
			[SMH_S.PRAYER_BEADS_BLESSING] = 190,
			[SMH_S.EPHEMERAL_POWER] = 175,
			[SMH_S.HEALING_OF_THE_AGES] = 350,
			[SMH_S.ESSENCE_OF_SAPPHIRON] = 130,
			[SMH_S.THE_EYE_OF_THE_DEAD] = 450,
			[SMH_S.ASCENDANCE] = 75,
			[SMH_S.CRUSADERS_WRATH] = 95,
			[SMH_S.CHROMATIC_INFUSION] = 190,
			[SMH_S.THE_FURIOUS_STORM] = 95,
			[SMH_S.TOASTED_SMORC] = 44,
			[SMH_S.SWEET_SURPRISE] = 44,
			[SMH_S.POWER_OF_THE_GUARDIAN] = 62,
			[SMH_S.NATURE_ALIGNED] = 250,
			[SMH_S.ADEPTS_ELIXIR] = 24,
			[SMH_S.ARCANE_ENERGY] = 220,
			[SMH_S.ARCANUM_OF_FOCUS] = 8,
			[SMH_S.ATIESH_PRIEST_DAMAGE_HEALING] = 300,
			[SMH_S.BLESSING_OF_RIGHTEOUSNESS] = 260,
			[SMH_S.CALL_OF_THE_NEXUS] = 225,
			[SMH_S.CENARION_BLESSING] = 175,
			[SMH_S.ELIXIR_OF_HEALING_POWER] = 50,
			[SMH_S.EMPOWER_CHAMPION] = 100,
			[SMH_S.ESSENCE_OF_THE_MARTYR] = 288,
			[SMH_S.FOCUSED_POWER] = 104,
			[SMH_S.HEALING_POWER] = 238,
			[SMH_S.INSIGHT] = 110,
			[SMH_S.LIVING_RUBY_SERPENT] = 150,
			[SMH_S.POWER_SIPHON] = -200,
			[SMH_S.RUBY_SERPENT] = 100,
			[SMH_S.SADIST] = 100,
			[SMH_S.SLYVAN_BLESSING] = 280,
			[SMH_S.SPELL_POWER] = 191,
			[SMH_S.THE_ARCANISTS_STONE] = 167,
			[SMH_S.POWER_OF_PRAYER] = 213,
			},
		buffTargetHealingEffects = {
			[SMH_S.ARCING_SMASH] = -50,
			[SMH_S.BANE_OF_INFINITY] = 50,
			[SMH_S.CHAINS_OF_KELTHUZAD] = 500,
			[SMH_S.CURSE_OF_INFINITY] = 50,
			[SMH_S.DEATHBLOW] = -50,
			[SMH_S.DEMOLISH] = -75,
			[SMH_S.EBON_POISON] = -25,
			[SMH_S.ENFEEBLE] = -100,
			[SMH_S.FEL_ARMOR] = 20,
			[SMH_S.FILET] = -50,
			[SMH_S.MAGMA_THROWERS_CURSE] = -50,
			[SMH_S.NETHER_PORTAL_DOMINANCE] = -1,
			[SMH_S.SOLAR_STRIKE] = -50,
			[SMH_S.SOUL_STRIKE] = -50,
			[SMH_S.WOUND_POISON] = -10,
			[SMH_S.AURA_OF_SUFFERING] = -100,
			[SMH_S.AURA_OF_DESIRE] = 100,
			[SMH_S.CURSE_OF_THE_BLEEDING_HOLLOW] = 15,
			[SMH_S.CURSE_OF_THE_DEADWOOD] = -50,
			[SMH_S.MORTAL_WOUND] = -10,
			[SMH_S.VEIL_OF_SHADOW] = -75,
			[SMH_S.BLOOD_FURY] = -50,
			[SMH_S.BROOD_AFFLICTION_GREEN] = -50,
			[SMH_S.MORTAL_STRIKE] = -50,
			[SMH_S.MORTAL_CLEAVE] = -50,
			[SMH_S.GEHENNAS_CURSE] = -75,
			[SMH_S.HEX_OF_WEAKNESS] = -20,
			[SMH_S.NECROTIC_POISON] = -90,
			},
		cachedSpellBookData = { },
		castingTimeReadProperly = false,	-- used to indicate whether the casting time being read from tooltips has been modified by a buff to make casts instant
		class = "None",				-- class of the player.  used to determine whether we should enable the addon.
		classSpellsOfInterest = {
			[SMH_E.DRUID] = { SMH_S.HEALING_TOUCH, SMH_S.LIFEBLOOM, SMH_S.REGROWTH, SMH_S.REJUVENATION },
			[SMH_E.PALADIN] = { SMH_S.FLASH_OF_LIGHT, SMH_S.HOLY_LIGHT },
			[SMH_E.PRIEST] = { SMH_S.BINDING_HEAL, SMH_S.FLASH_HEAL, SMH_S.FAKE_COMBINED_HEAL, SMH_S.LESSER_HEAL, SMH_S.RENEW },
			[SMH_E.SHAMAN] = { SMH_S.CHAIN_HEAL, SMH_S.HEALING_WAVE, SMH_S.LESSER_HEALING_WAVE },
			},
		classSpellsOfInterestMaxCount = 5,
		classesPermittedToUseAddon = {
			[SMH_E.DRUID] = true,
			[SMH_E.PALADIN] = true,
			[SMH_E.PRIEST] = true,
			[SMH_E.SHAMAN] = true,
			},
		currentOtherHeals = { },		-- heals by other users of this addon that are not yet complete
		currentOtherHealsByCharacter = { },	-- above heals sorted by character name
		debuffsCancellingCasting = { [SMH_S.BROOD_POWER_BRONZE] = true, [SMH_S.CALL_OF_THE_FOREST] = true, [SMH_S.CORRUPTED_FEAR] = true, [SMH_S.FEAR] = true,
						[SMH_S.FELL_CURSE_EFFECT] = true, [SMH_S.PESTER_EFFECT] = true, [SMH_S.SMOTHER] = true, [SMH_S.STONE_SLUMBER] = true, 
						[SMH_S.TERRIFYING_SCREECH] = true, },
		defaultCastTimes = {
			[SMH_S.HEALING_TOUCH] = { 1.5, 2, 2.5, 3, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5, 3.5 },
			[SMH_S.LIFEBLOOM] = { 0 },
			[SMH_S.REGROWTH] = { 2, 2, 2, 2, 2, 2, 2, 2, 2, 2 },
			[SMH_S.REJUVENATION] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
			[SMH_S.HOLY_LIGHT] = { 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 2.5 },
			[SMH_S.FLASH_OF_LIGHT] = { 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5 },
			[SMH_S.BINDING_HEAL] = { 1.5 },
			[SMH_S.LESSER_HEAL] = { 1.5, 2, 2.5 },
			[SMH_S.FAKE_COMBINED_HEAL] = { 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
			[SMH_S.HEAL] = { 3, 3, 3, 3 },
			[SMH_S.GREATER_HEAL] = { 3, 3, 3, 3, 3, 3, 3 },
			[SMH_S.RENEW] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
			[SMH_S.FLASH_HEAL] = { 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5 },
			[SMH_S.CHAIN_HEAL] = { 2.5, 2.5, 2.5, 2.5, 2.5 },
			[SMH_S.HEALING_WAVE] = { 1.5, 2, 2.5, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
			[SMH_S.LESSER_HEALING_WAVE] = { 1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 1.5 },
			},
		equipmentSlots = { "Head", "Neck", "Shoulder", "Back", "Chest", "Shirt", "Tabard", "Wrist", "Hands", "Waist", "Legs", "Feet", "Finger0", "Finger1", "Trinket0", "Trinket1", "MainHand", "SecondaryHand", "Ranged" },
		eventDispatch = {			-- table of functions whose keys are the events
			['CHARACTER_POINTS_CHANGED'] = function(self, arg1)
				--	arg1 tells us the delta in talent points
				if (arg1 == -1) then	-- a talent is learned
					self:CacheSpellBookData()	-- because spell tooltips are modified based on talents
					self:ProcessTalents()
				end
				if arg1 == 1 then	-- a level is gained
					self:UpdateTalentInformation()
				end
			end,
			['CHAT_MSG_ADDON'] = function(self, arg1, arg2, arg3, arg4)
				if arg1 == "SMH" then
					self:ChannelMessageProcess(arg2, arg4)
				end
			end,
			['LEARNED_SPELL_IN_TAB'] = function(self)
				self:CacheSpellBookData()
			end,
			['PLAYER_AURAS_CHANGED'] = function(self)
				self:UpdateTalentInformation()
				if not self.castingTimeReadProperly then
					self:CacheSpellBookData()
				end
				self:PlusHealingFromBuffs()
			end,
			['PLAYER_ENTERING_WORLD'] = function(self)
				self:ProcessTalents()
				self:UpdateTalentInformation()
				self:CacheSpellBookData()
				self:CacheEquipmentData()
				self:PlusHealingFromBuffs()
				this:RegisterEvent("CHARACTER_POINTS_CHANGED")
				this:RegisterEvent("CHAT_MSG_ADDON")
				this:RegisterEvent("LEARNED_SPELL_IN_TAB")
				this:RegisterEvent("PLAYER_AURAS_CHANGED")
				this:RegisterEvent("UNIT_INVENTORY_CHANGED")
			end,
			['UNIT_INVENTORY_CHANGED'] = function(self, arg1)
				--	arg1 tells us what unit had the inventory change
				if (arg1 == 'player') then
					self:CacheEquipmentData()
					self:UpdateTalentInformation()
				end
			end,
			['VARIABLES_LOADED'] = function(self)
				self:Init()
			end,
			},
		extraHealingAmount = 0,			-- amount of extra healing that is done by equipment
		extraHealingAmountPossible = 0,		-- amount of extra healing that is done by equipment if all were not broken
		gettingRegrowthSoon = { },		-- used to record who is getting a Regrowth so Druids do not attempt to cast it on them
		hasBenefitDreamwalker = false,		-- used to indicate whether the Dreamwalker set bonus for healing is present
		hasBenefitStormrage = false,		-- used to indicate whether the Stormrage set bonus for healing is present
		hasBenefitTranscendence = false,	-- used to indicate whether the Transcendence set bonus for healing is present
		hasBrokenPlusHealingItems = false,	-- indicates whether there are +healing items that are broken and therefore not included in the total
		hasItemHaroldsRejuvenatingBroach = 0,	-- used to indicate the number of +healing points for Rejuvenation the item gives
		hasItemIdolOfHealth = 0,		-- Healing Touch
		hasItemIdolOfPerseverance = 0,		-- Lifebloom
		hasItemIdolOfRejuvenation = 0,		-- used to indicate the number of +healing points for Rejuvenation the item gives
		hasItemIdolOfTenacity = 0,		-- Lifebloom
		hasItemIdolOfTheAvianHeart = 0,		-- Healing Touch
		hasItemIdolOfTheCrescentGoddess = 0,	-- Regrowth mana cost reduced
		hasItemIdolOfTheEmeraldQueen = 0,	-- Lifebloom
		hasItemIdolOfTheRavenGoddess = 0,	-- Tree Form extra healing
		hasItemLibramOfDivinity = 0,		-- used to indicate the number of +healing points for Flash of Light the item gives
		hasItemLibramOfLight = 0,		-- used to indicate the number of +healing points for Flash of Light the item gives
		hasItemLibramOfSoulsRedeemed = 0,	-- used to indicate the number of +healing points for Flash of Light and Holy light the item gives with Blessing of Light
		hasItemLibramOfTheLightbringer = 0,	-- used to indicate the number of +healing points for Holy Light
		hasItemTotemOfLife = 0,			-- used to indicate the number of +healing points for Lesser Healing Wave the item gives
		hasItemTotemOfSustaining = 0,		-- used to indicate the number of +healing points for Lesser Healing Wave the item gives
		healingEfficiency = 100,		-- used to indicate how efficient a heal will be.  mostly target debuffs lower this.
		levelLearned = {
			[SMH_S.HEALING_TOUCH] = { 1, 8, 14, 20, 26, 32, 38, 44, 50, 56, 60, 62, 68 },
			[SMH_S.LIFEBLOOM] = { 64 },
			[SMH_S.REGROWTH] = { 12, 18, 24, 30, 36, 42, 48, 54, 60, 66 },
			[SMH_S.REJUVENATION] = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58, 60, 64, 70 },
			[SMH_S.HOLY_LIGHT] = { 1, 6, 14, 22, 30, 38, 46, 54, 60, 62, 70 },
			[SMH_S.FLASH_OF_LIGHT] = { 20, 26, 34, 42, 50, 58, 66 },
			[SMH_S.BINDING_HEAL] = { 64 },
			[SMH_S.LESSER_HEAL] = { 1, 4, 10 },
			[SMH_S.FAKE_COMBINED_HEAL] = { 16, 22, 28, 34, 40, 46, 52, 58, 60, 63, 68 },
			[SMH_S.RENEW] = { 8, 14, 20, 26, 32, 38, 44, 50, 56, 60, 65, 70 },
			[SMH_S.FLASH_HEAL] = { 20, 26, 32, 38, 44, 50, 56, 61, 67 },
			[SMH_S.CHAIN_HEAL] = { 40, 46, 54, 61, 68 },
			[SMH_S.HEALING_WAVE] = { 1, 6, 12, 18, 24, 32, 40, 48, 56, 60, 63, 70 },
			[SMH_S.LESSER_HEALING_WAVE] = { 20, 28, 36, 44, 52, 60, 66 },
			},
		maxClassHealth = { [SMH_S.DRUID] = 3500, [SMH_S.HUNTER] = 3500, [SMH_S.MAGE] = 2350, [SMH_S.PALADIN] = 4700, [SMH_S.PRIEST] = 2350, [SMH_S.ROGUE] = 3500, [SMH_S.SHAMAN] = 3500, [SMH_S.WARLOCK] = 2350, [SMH_S.WARRIOR] = 4700 },
		maximumPercentageHealthAllowed = 300,
		modification = {
			[SMH_S.HOLY_GUIDANCE] = 0,
			[SMH_S.LUNAR_GUIDANCE] = 0,
			[SMH_S.NATURES_BLESSING] = 0,
			[SMH_S.NURTURING_INSTINCT] = 0,
			[SMH_S.SPIRITUAL_GUIDANCE] = 0,
			},
		name = "None",				-- name of the player.  used in determining raid group.
		notificationTable = { },		-- an array whose keys are notification names, and whose values are arrays of objects and methods
		oneUpNumber = 0,			-- used to identify casts
		--	This table provides the allowed percentage of +healing gear based on the default casting
		--	time of the spell, its healing duration for HOT spells, and the level learned.
		--	The penalties that have been precomputed include the following:
		--	For each level under 20 at which the spell is learned the penalty is 3.5%
		--	For a pure HOT spell, the allowed percentage is its duration / 15 seconds.
		--	For Regrowth, only 50% of the penalty for cast time is considered because of split HOT healing.
		--	For a non-pure HOT spell, the allowed percentage is the standard cast time compared against 3.5
		--	seconds.  Spells under 1.5 seconds (like instants) are considered 1.5 seconds.
		--	Lifebloom is instant cast, split HOT with only 7 seconds HOT duration.  Therefore, 50% of HOT
		--	of 7 / 15 seconds, plus 50% of instant (1.5 / 3.5) seconds.
		plusHealingAllowedPecentages = {
			[SMH_S.HEALING_TOUCH] = { 0, .1214, .4893, .8571, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
			[SMH_S.LIFEBLOOM] = { .4476 },
			[SMH_S.REGROWTH] = { .4857, .7107, .7857, .7857, .7857, .7857, .7857, .7857, .7857, .7857 },
			[SMH_S.REJUVENATION] = { .2, .425, .65, .8, .8, .8, .8, .8, .8, .8, .8, .8, .8 },
			[SMH_S.HOLY_LIGHT] = { .0018, .1893, .4893, .7143, .7143, .7143, .7143, .7143, .7143, .7143, .7143 },
			[SMH_S.FLASH_OF_LIGHT] = { .4286, .4286, .4286, .4286, .4286, .4286, .4286 },
			[SMH_S.BINDING_HEAL] = { .4286 },
			[SMH_S.LESSER_HEAL] = { 0, 0, .3393 },
			[SMH_S.FAKE_COMBINED_HEAL] = { .7171, .8571, .8571, .8571, .8571, .8571, .8571, .8571, .8571, .8571, .8571 },
			[SMH_S.RENEW] = { .58, .79, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
			[SMH_S.FLASH_HEAL] = { .4286, .4286, .4286, .4286, .4286, .4286, .4286, .4286, .4286 },
			[SMH_S.CHAIN_HEAL] = { .7143, .7143, .7143, .7143, .7143 },
			[SMH_S.HEALING_WAVE] = { 0, .0464, .4143, .7821, .8571, .8571, .8571, .8571, .8571, .8571, .8571, .8571 },
			[SMH_S.LESSER_HEALING_WAVE] = { .4286, .4286, .4286, .4286, .4286, .4286, .4286 },
			},
		plusHealingFromBuffs = 0,		-- +healing because of buffs on the player
		race,					-- this will be in English and is used for racial specific ability detection
		spellParsersMaxHeal = {
			[SMH_E.DRUID] = { [SMH_S.HEALING_TOUCH] = function(a) return SMH:PISecond(a) end, [SMH_S.LIFEBLOOM] = function(a) return SMH:PIFirstThird(a) end, [SMH_S.REJUVENATION] = function(a) return SMH:PIFirst(a) end, [SMH_S.REGROWTH] = function(a) return SMH:PISecondThird(a) end },
			[SMH_E.PALADIN] = { [SMH_S.HOLY_LIGHT] = function(a) return SMH:PISecond(a) end, [SMH_S.FLASH_OF_LIGHT] = function(a) return SMH:PISecond(a) end },
			[SMH_E.PRIEST] = { [SMH_S.BINDING_HEAL] = function(a) return SMH:PISecond(a) end, [SMH_S.LESSER_HEAL] = function(a) return SMH:PISecond(a) end, [SMH_S.HEAL] = function(a) return SMH:PISecond(a) end, [SMH_S.GREATER_HEAL] = function(a) return SMH:PISecond(a) end, [SMH_S.RENEW] = function(a) return SMH:PIFirst(a) end, [SMH_S.FLASH_HEAL] = function(a) return SMH:PISecond(a) end },
			[SMH_E.SHAMAN] = { [SMH_S.CHAIN_HEAL] = function(a) return SMH:PISecond(a) end, [SMH_S.HEALING_WAVE] = function(a) return SMH:PISecond(a) end, [SMH_S.LESSER_HEALING_WAVE] = function(a) return SMH:PISecond(a) end },
			},
		spellParsersMinHeal = {
			[SMH_E.DRUID] = { [SMH_S.HEALING_TOUCH] = function(a) return SMH:PIFirst(a) end, [SMH_S.LIFEBLOOM] = function(a) return SMH:PIFirstThird(a) end, [SMH_S.REJUVENATION] = function(a) return SMH:PIFirst(a) end, [SMH_S.REGROWTH] = function(a) return SMH:PIFirstThird(a) end, },
			[SMH_E.PALADIN] = { [SMH_S.HOLY_LIGHT] = function(a) return SMH:PIFirst(a) end, [SMH_S.FLASH_OF_LIGHT] = function(a) return SMH:PIFirst(a) end },
			[SMH_E.PRIEST] = { [SMH_S.BINDING_HEAL] = function(a) return SMH:PIFirst(a) end, [SMH_S.LESSER_HEAL] = function(a) return SMH:PIFirst(a) end, [SMH_S.HEAL] = function(a) return SMH:PIFirst(a) end, [SMH_S.GREATER_HEAL] = function(a) return SMH:PIFirst(a) end, [SMH_S.RENEW] = function(a) return SMH:PIFirst(a) end, [SMH_S.FLASH_HEAL] = function(a) return SMH:PIFirst(a) end },
			[SMH_E.SHAMAN] = { [SMH_S.CHAIN_HEAL] = function(a) return SMH:PIFirst(a) end, [SMH_S.HEALING_WAVE] = function(a) return SMH:PIFirst(a) end, [SMH_S.LESSER_HEALING_WAVE] = function(a) return SMH:PIFirst(a) end },
			},
		talentRank = {},			-- table of ranks of specific talents, indexed by name
		talentsOfInterest = {},			-- populated based on class so we only search for those needed

		BurningCrusadePlusHealingPercentage = function(self, spellName, spellRank, playerLevel)
			local retval = 0
			local levelToCompare = playerLevel or UnitLevel('player')

			if self.levelLearned[spellName][spellRank] then
				local levelOfSpellForBC
				local nextLevel = self.levelLearned[spellName][spellRank + 1]
				if nil ~= nextLevel then
					levelOfSpellForBC = nextLevel + 5
				else
					levelOfSpellForBC = 70
				end
				retval = levelOfSpellForBC / levelToCompare
				if retval > 1 then retval = 1 end
				if levelToCompare < self.levelLearned[spellName][spellRank] then retval = 0 end
			end

			return retval
		end,

		--	The purpose of this function is to look through all the equipped items
		--	and determine how much extra healing is provided by them.
		CacheEquipmentData = function(self)
			local oldExtraHealingAmount = self.extraHealingAmount
			local oldExtraHealingAmountPossible = self.extraHealingAmountPossible
			self.extraHealingAmount = 0
			self.extraHealingAmountPossible = 0
			self.hasItemHaroldsRejuvenatingBroach = 0
			self.hasItemIdolOfHealth = 0
			self.hasItemIdolOfPerseverance = 0
			self.hasItemIdolOfRejuvenation = 0
			self.hasItemIdolOfTenacity = 0
			self.hasItemIdolOfTheAvianHeart = 0
			self.hasItemIdolOfTheCrescentGoddess = 0
			self.hasItemIdolOfTheEmeraldQueen = 0
			self.hasItemIdolOfTheRavenGoddess = 0
			self.hasItemTotemOfLife = 0
			self.hasItemTotemOfSustaining = 0
			self.hasItemLibramOfDivinity = 0
			self.hasItemLibramOfLight = 0
			self.hasItemLibramOfSoulsRedeemed = 0
			self.hasItemLibramOfTheLightbringer = 0
			self.hasBenefitDreamwalker = false
			self.hasBenefitStormrage = false
			self.hasBenefitTranscendence = false
			self.hasBrokenPlusHealingItems = false
			local _SMH_setNames = { }
			local canLookForSetPatterns, itemIsBroken
			local slot

			for i, slot in ipairs(self.equipmentSlots) do
				com_mithrandir_smhTooltip:ClearLines()
				local id, _ = GetInventorySlotInfo(slot .. "Slot")
				local item, _, _ = com_mithrandir_smhTooltip:SetInventoryItem('player', id)
				if item then
					canLookForSetPatterns = false
					itemIsBroken = false
					itemPlusHealingValue = 0
					local lineCount = com_mithrandir_smhTooltip:NumLines()
					for j = 1, lineCount, 1 do
						local text = getglobal("com_mithrandir_smhTooltipTextLeft" .. j)
						if text then
							local line = text:GetText()

							--	First we look for specific items of interest
							if j == 1 then
								if line == SMH_S.ITEM_HAROLDS_REJUVENATING_BROACH then
									self.hasItemHaroldsRejuvenatingBroach = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_HEALTH then
									self.hasItemIdolOfHealth = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_PERSEVERANCE then
									self.hasItemIdolOfPerseverance = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_REJUVENATION then
									self.hasItemIdolOfRejuvenation = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_TENACITY then
									self.hasItemIdolOfTenacity = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_THE_AVIAN_HEART then
									self.hasItemIdolOfTheAvianHeart = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_THE_CRESCENT_GODDESS then
									self.hasItemIdolOfTheCrescentGoddess = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_THE_EMERALD_QUEEN then
									self.hasItemIdolOfTheEmeraldQueen = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_IDOL_OF_THE_RAVEN_GODDESS then
									self.hasItemIdolOfTheRavenGoddess = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_TOTEM_OF_LIFE then
									self.hasItemTotemOfLife = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_TOTEM_OF_SUSTAINING then
									self.hasItemTotemOfSustaining = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_LIBRAM_OF_DIVINITY then
									self.hasItemLibramOfDivinity = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_LIBRAM_OF_LIGHT then
									self.hasItemLibramOfLight = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_LIBRAM_OF_SOULS_REDEEMED then
									self.hasItemLibramOfSoulsRedeemed = self:NumberFromTooltipLine(lineCount); break
								elseif line == SMH_S.ITEM_LIBRAM_OF_THE_LIGHTBRINGER then
									self.hasItemLibramOfTheLightbringer = self:NumberFromTooltipLine(lineCount); break
								end
							end

							local match, _, setName = string.find(line, SMH_S.PATTERN_SET_NAME)
							if match ~= nil then
								if not _SMH_setNames[setName] then
									canLookForSetPatterns = true
									_SMH_setNames[setName] = true
								end
							end

							local match, where, current, total = string.find(line, SMH_S.PATTERN_DURABILITY)
							if match ~= nil then
								if tonumber(total) > 0 and tonumber(current) == 0 then itemIsBroken = true end
							end

							for k, pattern in ipairs(SMH_S.PATTERNS_HEALING) do
								local match, _, amount = string.find(line, pattern)
								if (match ~= nil) then
									self:OutputMessage("DEBUG", "On " .. slot .. " found " .. tonumber(amount) .. " healing.")
									itemPlusHealingValue = itemPlusHealingValue + tonumber(amount)
									break
								end
							end
							if canLookForSetPatterns then
								local possibleMatch
								possibleMatch, _ = string.find(line, SMH_S.PATTERN_STORMRAGE_SET)
								if possibleMatch ~= nil then
									self.hasBenefitStormrage = true
									self:OutputMessage("DEBUG", "Found Stormrage benefit.")
								else
									possibleMatch, _ = string.find(line, SMH_S.PATTERN_DREAMWALKER_SET)
									if possibleMatch ~= nil then
										self.hasBenefitDreamwalker = true
										self:OutputMessage("DEBUG", "Found Dreamwalker benefit.")
									else
										possibleMatch, _ = string.find(line, SMH_S.PATTERN_TRANSCENDENCE_SET)
										if possibleMatch ~= nil then
											self.hasBenefitTranscendence = true
											self:OutputMessage("DEBUG", "Found Transcendence benefit")
										else
											for k, pattern in ipairs(SMH_S.PATTERNS_SET_HEALING) do
												local match, _, amount = string.find(line, pattern)
												if match ~= nil then
													self:OutputMessage("DEBUG", "On "..slot.." found "..tonumber(amount).." healing as part of set.")
													itemPlusHealingValue = itemPlusHealingValue + tonumber(amount)
													break
												end
											end
										end
									end
								end

							end
						end
					end
					if itemPlusHealingValue > 0 then
						if itemIsBroken then
							self.hasBrokenPlusHealingItems = true
							self:OutputMessage("DEBUG", "Item in " .. slot .. " slot is broken, but should have provided " .. itemPlusHealingValue .. " +healing.")
							self.extraHealingAmountPossible = self.extraHealingAmountPossible + itemPlusHealingValue
						else
							self.extraHealingAmount = self.extraHealingAmount + itemPlusHealingValue
							self.extraHealingAmountPossible = self.extraHealingAmountPossible + itemPlusHealingValue
						end
					end
				end
			end

			if self.hasBenefitStormrage then
				--	The Stormrage benefit changes the duration of Rejuvenation so we change the time penalty for it
				self.plusHealingAllowedPecentages[SMH_S.REJUVENATION] = { .4, .625, .85, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 }
			else
				self.plusHealingAllowedPecentages[SMH_S.REJUVENATION] = { .2, .425, .65, .8, .8, .8, .8, .8, .8, .8, .8, .8, .8 }
			end

			self:OutputMessage("DEBUG", "Total extra healing found on equipment: ".. self.extraHealingAmount)

			if oldExtraHealingAmount ~= self.extraHealingAmount or oldExtraHealingAmountPossible ~= self.extraHealingAmountPossible then
				self:PostNotification(SMH_E.NOTIFICATION_EQUIPMENT_MODIFICATION_CHANGED)
			end

			return self.extraHealingAmount, self.extraHealingAmountPossible

		end,

		CacheSpellBookData = function(self)
			self:OutputMessage("DEBUG", "Caching spell book data.")
			-- clear the cache
			self.cachedSpellBookData = { }
			local localSpellsOfInterest = { }

			for i = 1, table.getn(self.classSpellsOfInterest[self.class]), 1 do
				self.cachedSpellBookData[self.classSpellsOfInterest[self.class][i]] = { }
				if self.classSpellsOfInterest[self.class][i] == SMH_S.FAKE_COMBINED_HEAL then
					localSpellsOfInterest[SMH_S.HEAL] = true
					localSpellsOfInterest[SMH_S.GREATER_HEAL] = true
				else
					localSpellsOfInterest[self.classSpellsOfInterest[self.class][i]] = true
				end
			end

			local playerBuffs = self:UnitBuffs('player')
			local casterHasNaturesSwiftness = (playerBuffs[SMH_S.NATURES_SWIFTNESS] ~= nil)
			local casterHasFocus = (playerBuffs[SMH_S.FOCUS] ~= nil)
			local casterHasBurningAdrenaline = (playerBuffs[SMH_S.BURNING_ADRENALINE] ~= nil)
			local casterHasInstantCast = casterHasNaturesSwiftness or casterHasFocus or casterHasBurningAdrenaline

			if casterHasInstantCast then self.castingTimeReadProperly = false else self.castingTimeReadProperly = true end

			local index = 1

			while (true) do
				local spellName, spellRankName = GetSpellName(index, SpellBookFrame.bookType)
				local maxHealAmount, minHealAmount

				-- If there is no spellName then we have run out of spells in the spell book
				if not spellName then break end

				-- If the spellName is in our table we have interest in processing it
				if (localSpellsOfInterest[spellName]) then
					local spellTexture = GetSpellTexture(index, SpellBookFrame.bookType)
					local _, _, spellRank = string.find(spellRankName, SMH_S.SPELLBOOK_RANK_FORMAT)
					if spellRank == nil then spellRank = "0" end
					local rankIndex = tonumber(spellRank)
					--	Now we look at the tooltips for each of the spells to pull out the mana cost
					--	and the amount of healing that can be done.  These values have been adjusted
					--	for the talents already.
					com_mithrandir_smhTooltip:ClearLines()
					com_mithrandir_smhTooltip:SetSpell(index, SpellBookFrame.bookType)
					local manaLine = com_mithrandir_smhTooltipTextLeft2:GetText()
					local manaCost
					if manaLine then
						_, _, manaCost = string.find(manaLine, SMH_S.SPELLBOOK_MANA_FORMAT)
						if manaCost == nil then manaCost = 0 end
					else
						manaCost = 0
					end
					local textLine = getglobal("com_mithrandir_smhTooltipTextLeft" .. com_mithrandir_smhTooltip:NumLines())
					local healLine = nil
					if textLine then
						healLine = textLine:GetText()
					else
						self:OutputMessage("INFO", "Problem parsing heal line because we are told there are " .. com_mithrandir_smhTooltip:NumLines() .. " but we cannot get that line.")
					end
					if healLine == nil then healLine = "" end
					if self.spellParsersMaxHeal[self.class][spellName] then
						maxHealAmount = self.spellParsersMaxHeal[self.class][spellName](healLine)
						if maxHealAmount == nil then
							self:OutputMessage("INFO", spellName.." (Rank "..spellRank..") cannot get max heal from: "..healLine)
							maxHealAmount = 0
						end
					else
						maxHealAmount = 0
					end
					if self.spellParsersMinHeal[self.class][spellName] then
						minHealAmount = self.spellParsersMinHeal[self.class][spellName](healLine)
						if minHealAmount == nil then
							self:OutputMessage("INFO", spellName.." (Rank "..spellRank..") cannot get min heal from: "..healLine)
							minHealAmount = 0
						end
					else
						minHealAmount = 0
					end

					local castingTime = 0
					if casterHasInstantCast then
						castingTime = self.defaultCastTimes[spellName][rankIndex]
					else
						if self.defaultCastTimes[spellName][rankIndex] ~= 0 then
							local castLine = com_mithrandir_smhTooltipTextLeft3:GetText()
							if castLine then
								_, _, castingTime = string.find(castLine, "^%s*([%d.]+)")
								--	It seems we actually get the event that causes us to attempt to parse
								--	the cast times but unfortunately the tooltip does not yet show the true
								--	values as they are still "Instant cast" and therefore, we need to still
								--	keep the data as invalid until we can once again reparse.
								if castingTime == nil then
									self.castingTimeReadProperly = false
									castingTime = self.defaultCastTimes[spellName][rankIndex]
									self:OutputMessage("DEBUG", "Could not find casting time for " .. spellName .. " (Rank " .. rankIndex .. ") from line: " .. castLine)
								end
							else
								castingTime = 0
							end
						else
							castingTime = 0
						end
					end

					local entry = { id = index, mana = tonumber(manaCost), maxHeal = maxHealAmount, minHeal = minHealAmount, castTime = tonumber(castingTime), texture = spellTexture }

					--	This little section of code takes the Priest spells Heal and Greater Heal
					--	and combines them into a fake spell that basically starts at Heal and then
					--	becomes Greater Heal when past the four Heals.  Basically this is done
					--	because Heal and Greater Heal seem to be the same spell in behavior and the
					--	amount they heal (and mana cost) goes up from the last Heal to the first
					--	Greater Heal.
					if spellName == SMH_S.HEAL then
						spellName = SMH_S.FAKE_COMBINED_HEAL
					elseif spellName == SMH_S.GREATER_HEAL then
						spellName = SMH_S.FAKE_COMBINED_HEAL
						rankIndex = rankIndex + 4
					end

					self.cachedSpellBookData[spellName][rankIndex] = entry
				end

				-- Next spell please
				index = index + 1
			end
			self:PostNotification(SMH_E.NOTIFICATION_CACHED_SPELLBOOK)
		end,

		ChannelMessageProcess = function(self, message, fromWhom)
			self:OutputMessage("DEBUG", "We have channel message from " .. fromWhom .. " that says: " .. message)
			if fromWhom ~= self.name then
				local match, _, spellNumber = string.find(message, "^SMH%:(%d+)%:Done$")
				if match ~= nil then
					--	We remove the entry in the table matching the spell
					self.currentOtherHeals[fromWhom .. "-" .. spellNumber] = nil
				else
					local match, _, spellNumber, theSpellName, theSpellRank, theRecipient, amount, theCastTime = string.find(message, "^SMH%:(%d+)%:([%w ]+)%:(%d+)%:(%w+)%:(%d+)%:([%d.]+)$")
					if match ~= nil then
						--	We add a new entry in the table for this spell
						self.currentOtherHeals[fromWhom .. "-" .. spellNumber] = { spellName = theSpellName, spellRank = theSpellRank, recipient = theRecipient, potentialHeal = amount, castTime = theCastTime, registered = GetTime() }
					end
				end
			end
		end,

		--	Copies the contents of a table into a new table
		CopyTable = function(self, someTable)
			local retval = { }
			for n in pairs(someTable) do
				if type(someTable[n]) == 'table' then
					retval[n] = self:CopyTable(someTable[n])
				else
					retval[n] = someTable[n]
				end
			end
			return retval
		end,

		--	This routine takes a groupInfo table and optional percentage to return
		--	a table of spell information indicating what spells the player should
		--	cast upon the unit specified in the groupinfo to bring it up to the
		--	percentage health.  When no health passed in, the user default is used.
		DetermineSpells = function(self, groupInfo, targetHealthPercentage, buffs)
			local retval = nil
			if groupInfo ~= nil and self:RealTableSize(groupInfo) == 4 then
				local currentConfiguration = nil
				if nil ~= SMH_Configuration then currentConfiguration = SMH_Configuration.profiles[SMH_Configuration.currentProfile] end
				local unitClass = UnitClass(groupInfo.id)
				local defaultPercentage = 100
				if currentConfiguration then
					defaultPercentage = currentConfiguration.classHealPercents[unitClass]
				end
				local desiredPercentage = targetHealthPercentage or defaultPercentage
				if desiredPercentage >= 0 and desiredPercentage <= self.maximumPercentageHealthAllowed then
					local playerBuffs = buffs or self:UnitBuffs('player')
					local targetBuffs
					if (groupInfo.id == 'player') then targetBuffs = playerBuffs else targetBuffs = self:UnitBuffs(groupInfo.id) end
					local percentageHealing = self:HealingEfficiency(playerBuffs, targetBuffs)
					-- If there is no percentage left for healing abaondon ship
					if percentageHealing <= 0 then return nil end

					retval = { }
					for i = 1, table.getn(self.classSpellsOfInterest[self.class]), 1 do
						retval[self.classSpellsOfInterest[self.class][i]] = { rank = 1, healAmount = 0, efficiency = percentageHealing }
					end

					local spellsToCheck = self.classSpellsOfInterest[self.class]
					for i = 1, table.getn(spellsToCheck), 1 do		-- loop through each of spellsToCheck
						spellName = spellsToCheck[i]

						-- first determine the level of spell needed based on healing required
						while (true) do
							spellInfo = self.cachedSpellBookData[spellName][retval[spellName].rank]
							if (spellInfo) then
								retval[spellName].healAmount = self:EstimatedHealingForTargetBuffs(playerBuffs, spellName, retval[spellName].rank, targetBuffs)

								if (((desiredPercentage / 100) * groupInfo.max) - groupInfo.current) < retval[spellName].healAmount then
									break
								end
								retval[spellName].rank = retval[spellName].rank + 1
							else
								retval[spellName].rank = retval[spellName].rank - 1
								break
							end
						end
					end

				else
					self:OutputMessage("DEBUG", "Target health percentage of " .. desiredPercentage .. " out of acceptable range 0 - " .. self.maximumPercentageHealthAllowed)
				end
			else
				self:OutputMessage("DEBUG", "Malformed groupInfo structure.")
			end
			return retval
		end,

		--	This convenience routine takes a unitId and optional percentage to return
		--	a table of spell information indicating what spells the player should cast
		--	upon the unitId to bring it up to percentage health.  With no health passed
		--	in, the user default is used instead.
		DetermineSpellsForUnit = function(self, unitId, targetHealthPercentage)
			local retval = nil
			if unitId ~= nil then
				local currentProfile = nil
				if nil ~= SMH_Configuration then currentProfile = SMH_Configuration.profiles[SMH_Configuration.currentProfile] end
				local unitClass = UnitClass(unitId)
				local defaultPercentage = 100
				local otherHealerConfidence = 100
				if currentProfile then
					defaultPercentage = currentProfile.classHealPercents[unitClass]
					otherHealerConfidence = currentProfile.otherHealerConfidence
				end
				local desiredPercentage = targetHealthPercentage or defaultPercentage
				if desiredPercentage >= 0 and desiredPercentage <= self.maximumPercentageHealthAllowed then
					local targetMaxHealth
					local targetCurrentHealth
					if self:UnitInGroup(unitId) then
						targetMaxHealth = UnitHealthMax(unitId)
						targetCurrentHealth = UnitHealth(unitId)
					else	-- we estimate the health information based on the level and class of the unit
						local slope = (self.maxClassHealth[unitClass] - 30) / 69
						targetMaxHealth = UnitLevel(unitId) * slope + 30
						targetCurrentHealth = targetMaxHealth * (UnitHealth(unitId) / 100)
					end

					-- now we determine if there are other heals that need to be handled when processing this recipient
					local characterName = UnitName(unitId)
					self:ProcessOtherHeals()
					local possibleHealingForUnit = self.currentOtherHealsByCharacter[characterName]
					if possibleHealingForUnit == nil then possibleHealingForUnit = 0 end
					targetCurrentHealth = targetCurrentHealth + possibleHealingForUnit * (otherHealerConfidence / 100)

					retval = self:DetermineSpells({ id = unitId, percent = targetCurrentHealth / targetMaxHealth, current = targetCurrentHealth, max = targetMaxHealth }, desiredPercentage)
				else
					self:OutputMessage("DEBUG", "Target health percentage of " .. desiredPercentage .. " out of acceptable range 0 - " .. self.maximumPercentageHealthAllowed)
				end
			else
				self:OutputMessage("DEBUG", "No unitId specified.")
			end
			return retval
		end,

		--	This routine provides the amount of healing done by the specified spellName at the
		--	specified spellRank under the influence of the specified buffs when targeting the
		--	specified unitId.  If the buffs is nil, the current player's buffs are used.  If
		--	the unitId is nil then any generic recipient is assumed to have no (de)buffs.
		EstimatedHealing = function(self, buffs, spellName, spellRank, unitId)
			local playerBuffs = buffs or self:UnitBuffs('player')
			local targetBuffs = nil
			if unitId == 'player' then targetBuffs = playerBuffs elseif unitId ~= nil then targetBuffs = self:UnitBuffs(unitId) end
			return self:EstimatedHealingForTargetBuffs(playerBuffs, spellName, spellRank, targetBuffs)
		end,

		EstimatedHealingForTargetBuffs = function(self, buffs, spellName, spellRank, targetBuffs)
			local retval = 0
			local currentConfiguration = nil
			if nil ~= SMH_Configuration then currentConfiguration = SMH_Configuration.profiles[SMH_Configuration.currentProfile] end
			local defaultHealingPercentage = 50
			if currentConfiguration then
				defaultHealingPercentage = currentConfiguration.healPercentage
			end
			local playerBuffs = buffs or self:UnitBuffs('player')
			local casterHasNaturesSwiftness = (playerBuffs[SMH_S.NATURES_SWIFTNESS] ~= nil)
			local casterHasFocus = (playerBuffs[SMH_S.FOCUS] ~= nil)
			local casterHasBurningAdrenaline = (playerBuffs[SMH_S.BURNING_ADRENALINE] ~= nil)
			local casterHasInstantCast = casterHasNaturesSwiftness or casterHasFocus or casterHasBurningAdrenaline

			local percentageHealing = self:HealingEfficiency(playerBuffs, targetBuffs)
			-- If there is no percentage left for healing abaondon ship
			if percentageHealing <= 0 then return 0 end
			local healAmount
			local spellInfo
			local temporaryExtraHealingAmount = 0	-- used for trinkets
			local plusHealAmountToUse
			local targetPlusHealingValue = 0

			local spellsToCheck = self.classSpellsOfInterest[self.class]
			local directGreaterHealAdd = 0
			local modifiedDirectGreaterHealAdd = false
			local temporaryPlusHealingBySpell = { }
			local plusHealingPercentageBySpell = { }
			-- Initialize our temporary array to 0 for each spell of interest
			for i = 1, table.getn(spellsToCheck), 1 do
				temporaryPlusHealingBySpell[spellsToCheck[i]] = 0
				if spellsToCheck[i] == SMH_S.FAKE_COMBINED_HEAL then
					plusHealingPercentageBySpell[SMH_S.GREATER_HEAL] = 1.0
					plusHealingPercentageBySpell[SMH_S.HEAL] = 1.0
				else
					plusHealingPercentageBySpell[spellsToCheck[i]] = 1.0
				end
			end

			--	Determine if we have modifications to the amount of +healing (usually based on talents)
			if self.class == SMH_E.DRUID then
				if nil ~= self.talentRank[SMH_S.EMPOWERED_REJUVENATION] then
					plusHealingPercentageBySpell[SMH_S.REGROWTH] = plusHealingPercentageBySpell[SMH_S.HEALING_TOUCH] + 0.04 * self.talentRank[SMH_S.EMPOWERED_REJUVENATION]
					plusHealingPercentageBySpell[SMH_S.REJUVENATION] = plusHealingPercentageBySpell[SMH_S.HEALING_TOUCH] + 0.04 * self.talentRank[SMH_S.EMPOWERED_REJUVENATION]
				end
				if nil ~= self.talentRank[SMH_S.EMPOWERED_TOUCH] then
					plusHealingPercentageBySpell[SMH_S.HEALING_TOUCH] = plusHealingPercentageBySpell[SMH_S.HEALING_TOUCH] + 0.10 * self.talentRank[SMH_S.EMPOWERED_TOUCH]
				end
			end
			if self.class == SMH_E.PRIEST then
				if nil ~= self.talentRank[SMH_S.EMPOWERED_HEALING] then
					plusHealingPercentageBySpell[SMH_S.FLASH_HEAL] = plusHealingPercentageBySpell[SMH_S.FLASH_HEAL] + 0.02 * self.talentRank[SMH_S.EMPOWERED_HEALING]
					plusHealingPercentageBySpell[SMH_S.GREATER_HEAL] = plusHealingPercentageBySpell[SMH_S.GREATER_HEAL] + 0.04 * self.talentRank[SMH_S.EMPOWERED_HEALING]
				end
			end

			--	Determine the spell specific +healing modifications and put them into the table.
			local possibleBuff
			if self.class == SMH_E.DRUID then
				temporaryPlusHealingBySpell[SMH_S.REJUVENATION] = temporaryPlusHealingBySpell[SMH_S.REJUVENATION] + self.hasItemHaroldsRejuvenatingBroach
				temporaryPlusHealingBySpell[SMH_S.REJUVENATION] = temporaryPlusHealingBySpell[SMH_S.REJUVENATION] + self.hasItemIdolOfRejuvenation
				temporaryPlusHealingBySpell[SMH_S.HEALING_TOUCH] = temporaryPlusHealingBySpell[SMH_S.HEALING_TOUCH] + self.hasItemIdolOfTheAvianHeart
				temporaryPlusHealingBySpell[SMH_S.HEALING_TOUCH] = temporaryPlusHealingBySpell[SMH_S.HEALING_TOUCH] + self.hasItemIdolOfHealth
				temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] = temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] + self.hasItemIdolOfTenacity
				temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] = temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] + self.hasItemIdolOfPerseverance
				temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] = temporaryPlusHealingBySpell[SMH_S.LIFEBLOOM] + self.hasItemIdolOfTheEmeraldQueen
				if self.hasBenefitDreamwalker then
					temporaryPlusHealingBySpell[SMH_S.REGROWTH] = temporaryPlusHealingBySpell[SMH_S.REGROWTH] + 350	-- hopefully this is truly the proper value
				end
			end
			if self.class == SMH_E.PALADIN then
				if nil ~= targetBuffs then
					possibleBuff= targetBuffs[SMH_S.GREATER_BLESSING_OF_LIGHT]
					if possibleBuff == nil then possibleBuff = targetBuffs[SMH_S.BLESSING_OF_LIGHT] end
					if possibleBuff ~= nil then
						temporaryPlusHealingBySpell[SMH_S.HOLY_LIGHT] = temporaryPlusHealingBySpell[SMH_S.HOLY_LIGHT] + possibleBuff[SMH_S.HOLY_LIGHT] + self.hasItemLibramOfSoulsRedeemed
						temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] = temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] + possibleBuff[SMH_S.FLASH_OF_LIGHT] + self.hasItemLibramOfSoulsRedeemed
					end
				end
				temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] = temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] + self.hasItemLibramOfDivinity
				temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] = temporaryPlusHealingBySpell[SMH_S.FLASH_OF_LIGHT] + self.hasItemLibramOfLight
				temporaryPlusHealingBySpell[SMH_S.HOLY_LIGHT] = temporaryPlusHealingBySpell[SMH_S.HOLY_LIGHT] + self.hasItemLibramOfTheLightbringer
			end
			if self.class == SMH_E.SHAMAN then
				temporaryPlusHealingBySpell[SMH_S.LESSER_HEALING_WAVE] = temporaryPlusHealingBySpell[SMH_S.LESSER_HEALING_WAVE] + self.hasItemTotemOfLife
				temporaryPlusHealingBySpell[SMH_S.LESSER_HEALING_WAVE] = temporaryPlusHealingBySpell[SMH_S.LESSER_HEALING_WAVE] + self.hasItemTotemOfSustaining
			end
			--	The Transcendence set can add healing value for the Greater Heal
			if self.class == SMH_E.PRIEST then
				if self.hasBenefitTranscendence then
					if self.cachedSpellBookData[SMH_S.RENEW] then spellInfo = self.cachedSpellBookData[SMH_S.RENEW][5] else spellInfo = nil end
					if spellInfo then
						directGreaterHealAdd = spellInfo.minHeal + ((spellInfo.maxHeal - spellInfo.minHeal) * defaultHealingPercentage / 100)
					end
				end
			end

			if nil ~= targetBuffs then
				-- The following two are mutually exclusive so the value of targetPlusHealingValue should be correct.
				possibleBuff = targetBuffs[SMH_S.AMPLIFY_MAGIC]
				if possibleBuff ~= nil then
					targetPlusHealingValue = targetPlusHealingValue + possibleBuff[SMH_S.AMPLIFY_MAGIC]
				end
				possibleBuff = targetBuffs[SMH_S.DAMPEN_MAGIC]
				if possibleBuff ~= nil then
					targetPlusHealingValue = targetPlusHealingValue  + possibleBuff[SMH_S.DAMPEN_MAGIC]
				end
			end

			modifiedDirectGreaterHealAdd = false

			spellInfo = self.cachedSpellBookData[spellName][spellRank]
			if spellInfo then

				--	Determine the base heal amount garnered from the
				--	tooltip for each spell.  This includes modifications
				--	from talents, but not from equipment.  It uses the
				--	technique the user wants for healing to determine
				--	where in the range of healing the target should be.
				healAmount = (spellInfo.minHeal + ((spellInfo.maxHeal - spellInfo.minHeal) * defaultHealingPercentage / 100))

				--	Deteremine which +healing buffs will still be active at the
				--	end of the normal time it takes to cast this buff.
				local castTime = spellInfo.castTime
				if casterHasInstantCast or castTime == nil then castTime = 0 end

				--	Check to see whether there are buffs that change the cast time (in seconds)
				castTime = castTime + self:SpellCastTimeModification(spellName, playerBuffs)
				if castTime < 0 then castTime = 0 end

				--	Check to see if there are buffs present that reduce the cast time
				if playerBuffs[SMH_S.BLOODLUST] ~= nil or playerBuffs[SMH_S.HEROISM] ~= nil then
					castTime = castTime * 0.70
				end

				temporaryExtraHealingAmount = self:PlusHealingFromBuffs(playerBuffs, castTime)
				if temporaryExtraHealingAmount == nil then temporaryExtraHealingAmount = 0 end
				temporaryExtraHealingAmount = temporaryExtraHealingAmount + targetPlusHealingValue
				if self:IsTreeOfLife(playerBuffs) then
					temporaryExtraHealingAmount = temporaryExtraHealingAmount + self.hasItemIdolOfTheRavenGoddess
				end

				plusHealAmountToUse = self.extraHealingAmount + temporaryExtraHealingAmount + temporaryPlusHealingBySpell[spellName] + self.modification[SMH_S.SPIRITUAL_GUIDANCE] + self.modification[SMH_S.NATURES_BLESSING] + self.modification[SMH_S.HOLY_GUIDANCE] + self.modification[SMH_S.LUNAR_GUIDANCE] + self.modification[SMH_S.NURTURING_INSTINCT]
				self:OutputMessage("DEBUG", spellName.." base: ".. healAmount .." raw +healing: ".. plusHealAmountToUse.." (E:".. self.extraHealingAmount.." T:".. temporaryExtraHealingAmount.." B:".. temporaryPlusHealingBySpell[spellName].." SG:".. self.modification[SMH_S.SPIRITUAL_GUIDANCE].." NB:"..self.modification[SMH_S.NATURES_BLESSING].." HG:"..self.modification[SMH_S.HOLY_GUIDANCE].." LG:"..self.modification[SMH_S.LUNAR_GUIDANCE].." NI:".. self.modification[SMH_S.NURTURING_INSTINCT]..")")

				burningCrusadePercentage = self:BurningCrusadePlusHealingPercentage(spellName, spellRank)

				-- For Greater Heal we add a little something on the rare chance we are supposed to do so.
				if spellName == SMH_S.FAKE_COMBINED_HEAL and spellRank > 4 then
					if not modifiedDirectGreaterHealAdd and self.hasBenefitTranscendence then
						directGreaterHealAdd = directGreaterHealAdd + (plusHealAmountToUse * self.plusHealingAllowedPecentages[SMH_S.RENEW][5] * self:BurningCrusadePlusHealingPercentage(SMH_S.RENEW, 5))
						modifiedDirectGreaterHealAdd = true
					end
					healAmount = healAmount + directGreaterHealAdd
				end

				--	Now modify the base amount by the amount of extra
				--	healing equipment can do.
				local spellNameToUse = spellName
				if spellName == SMH_S.FAKE_COMBINED_HEAL then
					if spellRank > 4 then
						spellNameToUse = SMH_S.GREATER_HEAL
					else
						spellNameToUse = SMH_S.HEAL
					end
				end
				if (plusHealAmountToUse > 0) then
					healAmount = healAmount + (self.plusHealingAllowedPecentages[spellName][spellRank] * plusHealAmountToUse * plusHealingPercentageBySpell[spellNameToUse] * burningCrusadePercentage)
				end
				retval = healAmount * (percentageHealing / 100)
				self:OutputMessage("DEBUG", spellName .. "(Rank " .. spellRank .. ") heals " .. healAmount .. " (including " .. self.plusHealingAllowedPecentages[spellName][spellRank] * plusHealAmountToUse .. " from a total of " .. plusHealAmountToUse .. " +healing, and TBC percentage of ".. burningCrusadePercentage * 100 .."%) with efficiency of "..percentageHealing.."%")

			end
			return retval
		end,

		-- This routines returns a value that indicates how efficient a heal will be when performed by the
		-- player with buffs against a target with targetBuffs
		HealingEfficiency = function(self, buffs, targetBuffs)
			local percentageHealing = 100
			local playerBuffs = buffs or self:UnitBuffs('player')
			local temp = { }
			--	Determine what buffs on the recipient and what on the caster influence the
			--	ability to heal.  Basically we modify the percentage of healing that can be
			--	done, starting with 100% and modify from there.
			if nil ~= targetBuffs then
				temp = { }
				for n in pairs(self.buffTargetHealingEffects) do table.insert(temp, n) end
				for i, n in ipairs(temp) do
					if targetBuffs[n] ~= nil then percentageHealing = percentageHealing + (self.buffTargetHealingEffects[n] * targetBuffs[n]["count"]) end
				end
			end
			temp = { }
			for n in pairs(self.buffCasterHealingEffects) do table.insert(temp, n) end
			for i, n in ipairs(temp) do
				if playerBuffs[n] ~= nil then percentageHealing = percentageHealing + (self.buffCasterHealingEffects[n] * playerBuffs[n]["count"]) end
			end

			if nil ~= playerBuffs[SMH_S.DIVINE_FAVOR] then
				percentageHealing = percentageHealing * 1.50	-- represents a critical
			end

			self.healingEfficiency = percentageHealing
			return percentageHealing
		end,

		Init = function(self)
			--	Ensure the tooltip is not messed up
			if not com_mithrandir_smhTooltip:IsOwned(UIParent) then
				com_mithrandir_smhTooltip:SetOwner(UIParent, "ANCHOR_NONE")
			end

			if self.class == SMH_E.DRUID then
				self.talentsOfInterest[SMH_S.EMPOWERED_REJUVENATION] = true
				self.talentsOfInterest[SMH_S.EMPOWERED_TOUCH] = true
				self.talentsOfInterest[SMH_S.LUNAR_GUIDANCE] = true
				self.talentsOfInterest[SMH_S.NURTURING_INSTINCT] = true
			end

			if self.class == SMH_E.PALADIN then
				self.talentsOfInterest[SMH_S.HOLY_GUIDANCE] = true
			end

			if self.class == SMH_E.PRIEST then
				self.talentsOfInterest[SMH_S.EMPOWERED_HEALING] = true
				self.talentsOfInterest[SMH_S.SPIRITUAL_GUIDANCE] = true
			end

			if self.class == SMH_E.SHAMAN then
				self.talentsOfInterest[SMH_S.NATURES_BLESSING] = true
			end

			this:RegisterEvent("PLAYER_ENTERING_WORLD")

		end,

		IsTreeOfLife = function(self, buffs)
			local retval = false
			local playerBuffs = buffs or self:UnitBuffs('player')
			if playerBuffs[SMH_S.TREE_OF_LIFE] ~= nil and playerBuffs[SMH_S.TREE_OF_LIFE]["form"] then retval = true end
			return retval
		end,

		NumberFromTooltipLine = function(self, desiredLineNumber)
			local retval = 0
			local theLine = getglobal("com_mithrandir_smhTooltipTextLeft" .. desiredLineNumber)

			if theLine then
				local theText = theLine:GetText()
				if theText then
					retval = self:PIFirst(theText)
				end
			end
			return retval
		end,

		OutputChannelMessage = function(self, addonName, spellName, spellRank, recipientName, estimatedHealAmount, spellCastTime)
			if recipientName == nil or estimatedHealAmount < 0 then return end
			local addonNameToUse = addonName or "SMH"
			local spellNameToUse = spellName or "NONE"
			local spellRankToUse = spellRank or 0
			local spellCastTimeToUse = spellCastTime or 0
			self.oneUpNumber = self.oneUpNumber + 1
			local message = "SMH:" .. self.oneUpNumber .. ":" .. spellNameToUse .. ":" .. spellRankToUse .. ":" .. recipientName .. ":" .. floor(estimatedHealAmount) .. ":" .. spellCastTimeToUse
			SendAddonMessage(addonNameToUse, message, "RAID")	-- Note that sending to RAID is fine because if not in a raid, but in a party it sends to party, and if not in a party nothing sent.
		end,

		OutputChannelMessageStop = function(self, addonName, how)
			local addonNameToUse = addonName or "SMH"
			local message = "SMH:" .. self.oneUpNumber .. ":Done"
			SendAddonMessage(addonNameToUse, message, "RAID")	-- Note that sending to RAID is fine because if not in a raid, but in a party it sends to party, and if not in a party nothing sent.
		end,

		--	The messageType indicates the "class" of message that we are sending.  This allows
		--	the various configuration options to control where the message should go.  The possible
		--	messageType values are DEBUG, TIER, ERROR, INFO and CAST.  A special value is used
		--	to determine what happens when an attempt at party is made but not in a party.
		--	The places to which a message can be sent are:
		--		0 -- no message sent, 1 -- chat log, 2 -- combat log, 3 -- error log, 4 -- party
		OutputMessage = function(self, messageType, message)
			local currentProfileName = nil
			if nil ~= SMH_Configuration then currentProfileName = SMH_Configuration.currentProfile end
			local messagePlace
			if nil == currentProfileName then
				if (messageType ~= 'DEBUG') then messagePlace = 1 else messagePlace = 0 end
			else
				messagePlace = SMH_Configuration.profiles[currentProfileName].messagePlaces[messageType]
			end

			if messagePlace == 4 and not UnitInParty('party1') then
				messagePlace = SMH_Configuration.profiles[currentProfileName].messagePlaces["CASTNOPARTY"]
			end

			if messagePlace == 1 and DEFAULT_CHAT_FRAME then
				DEFAULT_CHAT_FRAME:AddMessage(message)
			elseif messagePlace == 2 and ChatFrame2 then
				ChatFrame2:AddMessage(message)
			elseif messagePlace == 3 and UIErrorsFrame then
				UIErrorsFrame:AddMessage(message, 1.0, 1.0, 0.0, 1.0, UIERRORS_HOLD_TIME)
			elseif messagePlace == 4 then
				SendChatMessage(message, "PARTY")
			end
		end,

		ParseInteger = function(self, text, index)	-- taken from Healomatic
			local i = 1
			for integer in string.gmatch(text, '%d+') do
				if i == index then
					return tonumber(integer)
				end
				i = i + 1
			end
			return 0
		end,

		PIFirst = function(self, text)	return self:ParseInteger(text, 1)	end,

		PIFirstThird = function(self, text)	return self:ParseInteger(text, 1) + self:ParseInteger(text, 3)	end,

		PISecond = function(self, text)	return self:ParseInteger(text, 2)	end,

		PISecondThird = function(self, text)	return self:ParseInteger(text, 2) + self:ParseInteger(text, 3)	end,

		PlusHealingFromBuffs = function(self, buffs, castTime)
			local oldValue = self.plusHealingFromBuffs
			local retval = 0
			local playerBuffs = buffs or self:UnitBuffs('player')
			local spellCastTime = castTime or 0
			local temp = { }
			for n in pairs(self.buffCasterPlusHealingEffects) do table.insert(temp, n) end
			for i, n in ipairs(temp) do
				if playerBuffs[n] ~= nil and playerBuffs[n].lifetime > spellCastTime then
					retval = retval + self.buffCasterPlusHealingEffects[n]
				end
			end
			--	Since Unstable Power varies we need to treat it separately
			if playerBuffs[SMH_S.UNSTABLE_POWER] ~= nil and playerBuffs[SMH_S.UNSTABLE_POWER].lifetime > spellCastTime then
				retval = retval + playerBuffs[SMH_S.UNSTABLE_POWER][SMH_S.UNSTABLE_POWER]
			end
			self.plusHealingFromBuffs = retval
			if retval ~= oldValue then
				self:PostNotification(SMH_E.NOTIFICATION_BUFF_MODIFICATION_CHANGED)
			end
			return retval
		end,

		PostNotification = function(self, notificationName)
			local array = self.notificationTable[notificationName]
			if nil ~= array then
				for i, structure in ipairs(array) do
					structure.call()
				end
			end
		end,

		--	This routine goes through currentOtherHeals and maps valid entries
		--	into currentOtherHealsByCharacter so they can be easily accessed.
		--	Those entries older than 30 seconds are discarded.
		ProcessOtherHeals = function(self)
			local expired = { }
			local currentTime = GetTime()
			local temp
			local recipient
			local expiredCount

			self.gettingRegrowthSoon = { }
			self.currentOtherHealsByCharacter = { }
			for n in pairs(self.currentOtherHeals) do
				if currentTime - self.currentOtherHeals[n].registered > 30 then
					table.insert(expired, n)
				else
					recipient = self.currentOtherHeals[n].recipient
					temp = self.currentOtherHealsByCharacter[recipient]
					if temp == nil then temp = 0 end
					self.currentOtherHealsByCharacter[recipient] = temp + self.currentOtherHeals[n].potentialHeal
					if self.currentOtherHeals[n].spellName == SMH_S.REGROWTH then
						self.gettingRegrowthSoon[recipient] = true
					end
					self:OutputMessage("DEBUG", n .. " indicates " .. recipient .. " should be getting " .. self.currentOtherHeals[n].potentialHeal .. " health")
				end
			end
			expiredCount = table.getn(expired)
			for n = 1, expiredCount, 1 do
				self.currentOtherHeals[expired[n]] = nil
			end
		end,

		--	This function looks for specific talents for whatever reasons we need.
		ProcessTalents = function(self)
			self.talentRank = {}	-- clear talents to ensure no stale data
			local numberTalentTabs = GetNumTalentTabs()
			for i = 1, numberTalentTabs do
				local numberTalents = GetNumTalents(i)
				for j = 1, numberTalents do
					local name, iconTexture, tier, column, rank, maxRank = GetTalentInfo(i, j)
					if self.talentsOfInterest[name] and rank > 0 then
						self.talentRank[name] = rank
					end
				end
			end
		end,

		--	This function is used because there are times when I cannot
		--	call table.getn() on a table.  For example, if a table has
		--	got entries whose keys are the names of interest and the
		--	values are all true.  In this case table.getn() of that
		--	table returns 0.
		RealTableSize = function(self, someTable)
			local temp = { }

			for n in pairs(someTable) do table.insert(temp, n) end
			return table.getn(temp)
		end,

		RegisterForNotification = function(self, notificationName, method)
			local array = self.notificationTable[notificationName]
			if nil == array then
				self.notificationTable[notificationName] = { }
				array = self.notificationTable[notificationName]
			end
			local structure = { call = method }
			table.insert(array, structure)
		end,

		--	This returns a number against which the mana cost should be multiplied to
		--	determine the true mana cost.
		SpellCastManaModification = function(self, spellName, buffs)
			local retval = 1
			local playerBuffs = buffs or self:UnitBuffs('player')

			if nil ~= playerBuffs[SMH_S.DIVINE_ILLUMINATION] then
				retval = retval * 0.50
			end
			if nil ~= playerBuffs[SMH_S.NATURE_ALIGNED] then
				retval = retval * 1.20
			end
			if self:IsTreeOfLife(playerBuffs) then
				if spellName == SMH_S.REGROWTH or spellName == SMH_S.REJUVENATION or spellName == SMH_S.LIFEBLOOM then
					retval = retval * 0.80
				end
			end
			if spellName == SMH_S.REGROWTH then
				retval = retval - self.hasItemIdolOfTheCrescentGoddess
			end
			return retval
		end,

		--	This returns a number that should be added to the casting time to determine
		--	the actual casting time.
		SpellCastTimeModification = function(self, spellName, buffs)
			local retval = 0
			local playerBuffs = buffs or self:UnitBuffs('player')
			
			if nil ~= playerBuffs[SMH_S.NATURES_GRACE] then
				retval = retval - 0.5
			end
			if nil ~= playerBuffs[SMH_S.LIGHTS_GRACE] and spellName == SMH_S.HOLY_LIGHT then
				retval = retval - 0.5
			end
			return retval
		end,

		Tooltip_OnEvent = function(self)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9)
			end
		end,

		Tooltip_OnLoad = function(self)
			_, self.class = UnitClass('player')
			self.name = UnitName('player')
			if self.classesPermittedToUseAddon[self.class] then
				_, self.race = UnitRace('player')
				this:SetScript("OnEvent", function() self:Tooltip_OnEvent() end)
				this:RegisterEvent("VARIABLES_LOADED")
			end
		end,

		--	This returns a table containing keys of the buff names.
		UnitBuffs = function(self, unitName)
			local retval = { }
			local texture, appliedCount, buffname
			for i = 0, 40, 1 do
				com_mithrandir_smhTooltip:ClearLines()
				_, _, texture, appliedCount = UnitBuff(unitName, i)
				if texture then
					local realAppliedCount = tonumber(appliedCount)
					if realAppliedCount == 0 then realAppliedCount = 1 end	-- handles case when the buff is not stackable
					com_mithrandir_smhTooltip:SetUnitBuff(unitName, i)
					buffName = com_mithrandir_smhTooltipTextLeft1:GetText()
					if buffName then
						if nil == retval[buffName] then
							retval[buffName] = { ["count"] = realAppliedCount, ["lifetime"] = 0 }
						end
						if buffName == SMH_S.GREATER_BLESSING_OF_LIGHT or buffName == SMH_S.BLESSING_OF_LIGHT then
							local description = com_mithrandir_smhTooltipTextLeft2:GetText()
							local one = self:PIFirst(description)
							local two = self:PISecond(description)
							if one and two then
								retval[buffName][SMH_S.HOLY_LIGHT] = one
								retval[buffName][SMH_S.FLASH_OF_LIGHT] = two
							end
						end
						if buffName == SMH_S.UNSTABLE_POWER or buffName == SMH_S.AMPLIFY_MAGIC then
							local description = com_mithrandir_smhTooltipTextLeft2:GetText()
							local two = self:PISecond(description)
							if two then
								retval[buffName][buffName] = two
							end
						end

						if buffName == SMH_S.DAMPEN_MAGIC then
							local description = com_mithrandir_smhTooltipTextLeft2:GetText()
							local two = self:PISecond(description)
							if two then
								retval[buffName][buffName] = two * -1
							end
						end

						--	Determine if under influence of Tree of Life
						if buffName == SMH_S.TREE_OF_LIFE then
							local description = com_mithrandir_smhTooltipTextLeft2:GetText()
							if description then
								local match, _ = string.find(description, SMH_S.PATTERN_TREE_OF_LIFE_FORM)
								if match ~= nil then
									retval[buffName]["form"] = true
								else
									retval[buffName]["influence"] = true	-- really want to be able to know which Tree of Life if can have more than one
								end
							end
						end

						--	Now we record whether any of these buffs are ones indicating being mounted
						local description = com_mithrandir_smhTooltipTextLeft2:GetText()
						if description then
							local match, _, amount = string.find(description, SMH_S.PATTERN_MOUNT)
							if match ~= nil then
								retval[buffName]["mounted"] = true
							end
							local match, _, amount = string.find(description, SMH_S.PATTERN_FLYING_MOUNT)
							if match ~= nil then
								retval[buffName]["mounted"] = true
							end
						end

						--	Now we determine whether any of these buffs disable casting
						if self.buffsCancellingCasting[buffName] then
							retval[buffName]["castStopping"] = true
						end
					end

				end
				com_mithrandir_smhTooltip:ClearLines()
				_, _, texture, appliedCount = UnitDebuff(unitName, i)
				if texture then
					com_mithrandir_smhTooltip:SetUnitDebuff(unitName, i)
					local debuffName = com_mithrandir_smhTooltipTextLeft1:GetText()
					if debuffName then
						local realAppliedCount = tonumber(appliedCount)
						if realAppliedCount == 0 then realAppliedCount = 1 end
						retval[debuffName] = { ["count"] = realAppliedCount, ["lifetime"] = 0 }

						if self.debuffsCancellingCasting[debuffName] then
							retval[debuffName]["castStopping"] = true
						else
							--	Now check to see whether any of these are considered castStopping
							local lineCount = com_mithrandir_smhTooltip:NumLines()
							for j = 1, lineCount, 1 do
								local text = getglobal("com_mithrandir_smhTooltipTextLeft"..j)
								if text then
									local line = text:GetText()

									if line then
										for k, pattern in ipairs(SMH_S.PATTERNS_CAST_DISABLING) do
											local match, _ = string.find(line, pattern)
											if (match ~= nil) then
												retval[debuffName]["castStopping"] = true
												break
											end
										end
									end
								end
							end
						end
					end
				end
			end

			--	Get the time left remaining for each of the (de)buffs
			if unitName == 'player' then
				local buffIndex
				local timeRemaining
				local buffName

				for i = 0, 40, 1 do
					com_mithrandir_smhTooltip:ClearLines()
					buffIndex = GetPlayerBuff(i, "HELPFUL|HARMFUL|PASSIVE")
					if buffIndex >= 0 then
						com_mithrandir_smhTooltip:SetPlayerBuff(buffIndex)
						buffName = com_mithrandir_smhTooltipTextLeft1:GetText()
						if buffName and retval[buffName] then
							timeRemaining = GetPlayerBuffTimeLeft(buffIndex)
							retval[buffName]["lifetime"] = timeRemaining
						end
					end
				end
			end

			return retval
		end,

		--	This routine returns if the unitId is associated with the
		--	caster in a group (party or raid).
		UnitInGroup = function(self, unitId)
			local retval = (unitId == 'player' or unitId == 'pet')
			if not retval then retval = self:UnitInParty(unitId) end
			if not retval then retval = self:UnitInRaid(unitId) end
			return retval
		end,

		--	Note that UnitInParty will not return 1 if the unitId is
		--	a pet.  However, we want it to do so.  And now it appears
		--	as of patch 1.11.2 'player' is not considered part of the
		--	party either.  Starting in 1.12 we have new API to use.
		UnitInParty = function(self, unitId)
			return UnitPlayerOrPetInParty(unitId)
		end,

		--	Note that UnitInRaid will not return 1 if the unitId is
		--	a pet.  However, we want it to do so.
		UnitInRaid = function(self, unitId)
			return UnitPlayerOrPetInRaid(unitId)
		end,

		UpdateTalentStatData = function(self, statIndex, talentName, notificationName, percentage)
			local base, current, posBuff, negBuff = UnitStat('player', statIndex)
			local oldValue = self.modification[talentName]
			if nil ~= self.talentRank[talentName] then
				self.modification[talentName] = current * self.talentRank[talentName] * percentage
			else
				self.modification[talentName] = 0
			end
			if oldValue ~= self.modification[talentName] then
				self:PostNotification(notificationName)
			end
		end,

		UpdateTalentInformation = function(self)
			if self.talentRank[SMH_S.SPIRITUAL_GUIDANCE] ~= nil and self.talentRank[SMH_S.SPIRITUAL_GUIDANCE] > 0 then
				self:UpdateTalentStatData(5, SMH_S.SPIRITUAL_GUIDANCE, SMH_E.NOTIFICATION_SPIRITUAL_GUIDANCE_MODIFICATION_CHANGED, 0.05)
			end
			if self.talentRank[SMH_S.NATURES_BLESSING] ~= nil and self.talentRank[SMH_S.NATURES_BLESSING] > 0 then
				self:UpdateTalentStatData(4, SMH_S.NATURES_BLESSING, SMH_E.NOTIFICATION_NATURES_BLESSING_MODIFICATION_CHANGED, 0.10)
			end
			if self.talentRank[SMH_S.HOLY_GUIDANCE] ~= nil and self.talentRank[SMH_S.HOLY_GUIDANCE] > 0 then
				self:UpdateTalentStatData(4, SMH_S.HOLY_GUIDANCE, SMH_E.NOTIFICATION_HOLY_GUIDANCE_MODIFICATION_CHANGED, 0.07)
			end
			if self.talentRank[SMH_S.LUNAR_GUIDANCE] ~= nil and self.talentRank[SMH_S.LUNAR_GUIDANCE] > 0 then
				self:UpdateTalentStatData(4, SMH_S.LUNAR_GUIDANCE, SMH_E.NOTIFICATION_LUNAR_GUIDANCE_MODIFICATION_CHANGED, 0.0833)
			end
			if self.talentRank[SMH_S.NURTURING_INSTINCT] ~= nil and self.talentRank[SMH_S.NURTURING_INSTINCT] > 0 then
				self:UpdateTalentStatData(1, SMH_S.NURTURING_INSTINCT, SMH_E.NOTIFICATION_NURTURING_INSTINCT_MODIFICATION_CHANGED, 0.25)
			end
		end,

		}

end
